from collections import defaultdict

import trio
from trio.hazmat import ParkingLot as WaitQueue


def _ANY_TRANSITION(value, old_value):
    return True


class _Result:
    __slots__ = ['event', 'value']

    def __init__(self):
        self.event = WaitQueue()
        self.value = None


class AsyncValue:
    """
    Value wrapper offering the ability to wait for a value or transition.

    Synopsis:
        >>> a = AsyncValue(0)  # NOTE: can wrap any type (enum, tuple, ...)
        >>> ...
        >>> a.value = 5   # access underlying value
        >>> ...
        >>> # wait for value predicate
        >>> await a.wait_value(lambda v: v > 10)
        >>> ...
        >>> # wait for transition predicate (default: any)
        >>> await a.wait_transition(lambda v, old: v > 10 and old < 0)

    When using `wait_value()` and `wait_transition()`, note that the value may
    have changed again before the caller receives control.

    Performance note:  assignment to the `value` property has O(N) complexity,
    where N is the number of active wait predicates.
    """

    def __init__(self, value):
        self._value = value
        self._old = None  # NOTE: initial value is never exposed in API
        self._level_results = defaultdict(_Result)
        self._edge_results = defaultdict(_Result)

    def __repr__(self):
        return f"AsyncValue(value={self.value})"

    @property
    def value(self):
        """The wrapped value"""
        return self._value

    @value.setter
    def value(self, x):
        if self._value != x:
            old = self._old = self._value
            new = self._value = x
            for f, result in list(self._level_results.items()):
                if f(new):
                    result.value = new
                    result.event.unpark_all()
                    del self._level_results[f]
            for f, result in list(self._edge_results.items()):
                if f(new, old):
                    result.value = (new, old)
                    result.event.unpark_all()
                    del self._edge_results[f]

    async def wait_value(self, predicate):
        """
        Wait until given predicate f(value) is True.

        The predicate is tested immediately and, if false, whenever
        the `value` property changes.

        returns value which satisfied the predicate
        """
        if not predicate(self._value):
            result = self._level_results[predicate]
            await result.event.park()
            value = result.value
        else:
            value = self._value
            await trio.sleep(0)
        return value

    async def wait_transition(self, predicate=None):
        """
        Wait until given predicate f(value, old_value) is True.

        The predicate is tested whenever the `value` property changes.
        The default is `None`, which responds to any value change.

        returns (value, old_value) which satisfied the predicate
        """
        result = self._edge_results[predicate or _ANY_TRANSITION]
        await result.event.park()
        return result.value



import logging
from collections import defaultdict

import trio

logger = logging.getLogger(__name__)

RATE_MEASURE_PERIOD = .25


class TaskStats(trio.abc.Instrument):
    """Trio scheduler Instrument which logs various task stats at termination.

    Includes max task wait time, slowest task step, and highest task schedule
    rate.
    """
    def __init__(self, current_time=trio.current_time):
        super().__init__()
        self.current_time = current_time
        self.scheduled_start = {}
        self.max_wait = 0
        self.task_step_start = None
        self.max_task_step = (None, 0)
        self.schedule_counts = defaultdict(int)
        self.rate_start = 0
        self.max_schedule_rate = (None, 0)

    def task_scheduled(self, task):
        t = self.current_time()
        self.scheduled_start[task] = t
        if t - self.rate_start > RATE_MEASURE_PERIOD:
            name, count = max(self.schedule_counts.items(),
                              default=(None, 0),
                              key=lambda item: item[1])
            rate = count / RATE_MEASURE_PERIOD
            if rate > self.max_schedule_rate[1]:
                self.max_schedule_rate = (name, rate)
            self.rate_start = t
            self.schedule_counts.clear()
        self.schedule_counts[task.name] += 1

    def before_task_step(self, task):
        t = self.current_time()
        start = self.scheduled_start.pop(task, None)
        if start:
            dt = t - start
            self.max_wait = max(self.max_wait, dt)
            self.task_step_start = t

    def after_task_step(self, task):
        start = self.task_step_start
        if start:
            dt = self.current_time() - start
            if dt > self.max_task_step[1]:
                self.max_task_step = (task.name, dt)
            self.task_step_start = None

    def after_run(self):
        logger.info(f'max task wait time: {self.max_wait * 1000:.2f} ms')
        max_task_name, max_task_time = self.max_task_step
        logger.info(f'slowest task step: {max_task_time * 1000:.2f} ms by {max_task_name}')
        max_rate_name, max_rate = self.max_schedule_rate
        logger.info(f'max task schedule rate: {max_rate:.0f} Hz by {max_rate_name}')



import trio

async def periodic(period):
    """Yield `(elapsed_time, delta_time)` with an interval of `period` seconds.

    For example, to loop indefinitely with a period of 1 second, accounting
    for the time taken in the loop itself::

        async for _ in periodic(1):
            ...

    In the case of overrun, the next iteration begins immediately.

    On the first iteration, `delta_time` will be `None`.
    """
    t0 = trio.current_time()
    t_last = None
    while True:
        t_start = t0 if t_last is None else trio.current_time()
        yield t_start - t0, None if t_last is None else t_start - t_last
        user_elapsed = trio.current_time() - t_start
        await trio.sleep(max(0, period - user_elapsed))
        t_last = t_start



from ._version import __version__
from ._async_bool import AsyncBool
from ._async_dictionary import AsyncDictionary
from ._async_itertools import azip, azip_longest
from ._async_value import AsyncValue
from ._awaitables import wait_all, wait_any
from ._exceptions import defer_to_cancelled
from ._periodic import periodic
from ._repeated_event import UnqueuedRepeatedEvent, MailboxRepeatedEvent
from ._task_stats import TaskStats

def _metadata_fix():
    # don't do this for Sphinx case because it breaks "bysource" member ordering
    import sys
    if 'sphinx' in sys.modules:
        return

    for name, value in globals().items():
        if not name.startswith('_'):
            value.__module__ = __name__

_metadata_fix()



from contextlib import _GeneratorContextManager
from functools import wraps
from inspect import iscoroutinefunction
from typing import Type

import trio


class _AsyncFriendlyGeneratorContextManager(_GeneratorContextManager):
    """
    Hack contextlib._GeneratorContextManager so that resulting context
    manager can properly decorate async functions.
    """

    def __call__(self, func):
        if iscoroutinefunction(func):
            wraps(func)
            async def inner(*args, **kwargs):
                with self._recreate_cm():  # pylint: disable=not-context-manager
                    return await func(*args, **kwargs)
        else:
            wraps(func)
            def inner(*args, **kwargs):
                with self._recreate_cm():  # pylint: disable=not-context-manager
                    return func(*args, **kwargs)
        return inner


def async_friendly_contextmanager(func):
    """
    Equivalent to @contextmanager, except the resulting (non-async) context
    manager works correctly as a decorator on async functions.
    """
    @wraps(func)
    def helper(*args, **kwargs):
        return _AsyncFriendlyGeneratorContextManager(func, args, kwargs)
    return helper


@async_friendly_contextmanager
def defer_to_cancelled(*args: Type[Exception]):
    """Context manager which defers MultiError exceptions to Cancelled.

    In the scope of this context manager, any raised trio.MultiError exception
    which is a combination of the given exception types and trio.Cancelled
    will have the exception types filtered, leaving only a Cancelled exception.

    The intended use is where routine exceptions (e.g. which are part of an API)
    might occur simultaneously with Cancelled (e.g. when using move_on_after()).
    Without properly catching and filtering the resulting MultiError, an
    unhandled exception will occur.  Often what is desired in this case is
    for the Cancelled exception alone to propagate to the cancel scope.

    :param args:  One or more exception types which will defer to
        trio.Cancelled.  By default, all exception types will be filtered.

    Example::

        # If MultiError([Cancelled, Obstacle]) occurs, propagate only Cancelled
        # to the parent cancel scope.
        with defer_to_cancelled(Obstacle):
            try:
                # async call which may raise exception as part of API
                await advance(speed)
            except Obstacle:
                # handle API exception (unless Cancelled raised simultaneously)
                ...

    TODO: Support consolidation of simultaneous user API exceptions
        (i.e. MultiError without Cancelled).  This would work by prioritized list
        of exceptions to defer to.  E.g. given::

            [Cancelled, WheelObstruction, RangeObstruction]

        then::

            Cancelled + RangeObstruction => Cancelled
            WheelObstruction + RangeObstruction => WheelObstruction
    """
    try:
        yield
    except trio.MultiError as e:
        exceptions = e.exceptions
        if not any(isinstance(exc, trio.Cancelled) for exc in exceptions):
            raise
        if not args:
            raise trio.MultiError.filter(
                lambda exc: exc if isinstance(exc, trio.Cancelled) else None,
                e)
        raise trio.MultiError.filter(
            lambda exc: None if isinstance(exc, args) else exc,
            e)



from collections import defaultdict

import trio
from trio.hazmat import ParkingLot as WaitQueue

try:
    # work around numpy bool madness
    import numpy  # pylint: disable=import-error
    _BOOL_TYPES = (bool, numpy.bool_)
except ModuleNotFoundError:
    _BOOL_TYPES = (bool,)  # type: ignore

_BOOL_OR_NONE_TYPES = _BOOL_TYPES + (type(None), )


class AsyncBool:
    """Boolean wrapper offering the ability to wait for a value or transition.

    Synopsis:
        >>> a = AsyncBool()
        >>> ...
        >>> a.value = True   # access underlying value
        >>> ...
        >>> await a.wait_value(False)  # wait for a specific value
        >>> ...
        >>> await a.wait_transition()  # wait for a transition (default: any)

    When using `wait_value()` and `wait_transition()`, note that the value may
    have changed again before the caller receives control.
    """

    def __init__(self, value=False):
        if not isinstance(value, _BOOL_TYPES):
            raise TypeError
        self._value = value
        self._level_events = defaultdict(WaitQueue)
        self._edge_events = defaultdict(WaitQueue)

    def __repr__(self):
        return f"AsyncBool(value={self._value})"

    @property
    def value(self):
        """The wrapped value"""
        return self._value

    @value.setter
    def value(self, x):
        if not isinstance(x, _BOOL_TYPES):
            raise TypeError
        if self._value != x:
            self._value = x
            self._level_events[x].unpark_all()
            self._edge_events[x].unpark_all()
            self._edge_events[None].unpark_all()

    async def wait_value(self, value):
        """Wait until given value."""
        if not isinstance(value, _BOOL_TYPES):
            raise TypeError
        if value != self.value:
            await self._level_events[value].park()
        else:
            await trio.sleep(0)

    async def wait_transition(self, value=None):
        """Wait until transition to given value (default None which means any)."""
        if not isinstance(value, _BOOL_OR_NONE_TYPES):
            raise TypeError
        await self._edge_events[value].park()



from trio.hazmat import ParkingLot as WaitQueue


class UnqueuedRepeatedEvent:
    """An unqueued repeated event.

    A repeated event may be triggered multiple times, and has only one
    listener.  A call to set() is ignored if the listener is still processing
    the previous event (or there is no listener), and won't be queued.

    >>> event = UnqueuedRepeatedEvent()

    One task runs a listening loop, waiting for set():

    >>> async for _ in event:
    >>>    # do blocking work
    >>>    await trio.sleep(1)

    Another task triggers events:

    >>> event.set()  # trigger event
    >>> trio.sleep(0)  # event processing starts
    >>> event.set()  # ignored, because previous event is still being processed
    >>> trio.sleep(2)
    >>> event.set()  # trigger event
    """

    def __init__(self):
        self._wait_queue = WaitQueue()
        self._iteration_open = False

    def set(self):
        """Trigger event if there is a waiting listener."""
        self._wait_queue.unpark_all()

    def __aiter__(self):
        if self._iteration_open:
            raise RuntimeError(f'{self.__class__.__name__} can only have one listener')
        return self._listen()

    async def _listen(self):
        self._iteration_open = True
        try:
            while True:
                await self._wait_queue.park()
                yield
        finally:
            self._iteration_open = False


class MailboxRepeatedEvent:
    """A repeated event which queues up to one set() call.

    A repeated event may be triggered multiple times, and has only one
    listener.  Up to one set() call may be queued if the listener is
    still processing the previous event.

    It's often used for signaling that an out-of-band data location has
    changed (hence the name "mailbox").  If the data is a collection type
    (list, etc.) it's safe to use this class, but other types of data are
    subject to overrun.

    AVOID USING THIS CLASS FOR NON-COLLECTION DATA because data changes can
    be lost if the listener does not keep up.  Consider using a queue instead
    (see trio.open_memory_channel) so that there is back pressure ensuring
    that the data is received.

    >>> event = MailboxRepeatedEvent()
    >>> data = None

    One task runs a listening loop, waiting for set():

    >>> async for _ in event:
    >>>    # process data
    >>>    print(data)
    >>>    await trio.sleep(1)

    Another task triggers iteration:

    >>> data = 'foo'
    >>> event.set()  # trigger event
    >>> trio.sleep(0)  # event processing starts
    >>> data = 'bar'
    >>> event.set()  # trigger event (queued since previous event is being processed)
    """

    def __init__(self):
        self._requested = False
        self._wait_queue = WaitQueue()
        self._iteration_open = False

    def set(self):
        """Trigger event

        Up to one event may be queued if there is no waiting listener (i.e.
        no listener, or the listener is still processing the previous event).
        """
        self._requested = True
        self._wait_queue.unpark_all()

    def __aiter__(self):
        if self._iteration_open:
            raise RuntimeError(f'{self.__class__.__name__} can only have one listener')
        return self._listen()

    async def _listen(self):
        self._iteration_open = True
        try:
            while True:
                if not self._requested:
                    await self._wait_queue.park()
                self._requested = False
                yield
        finally:
            self._iteration_open = False



from typing import TypeVar, MutableMapping

import trio


KT = TypeVar('KT')
VT = TypeVar('VT')


class AsyncDictionary(MutableMapping[KT, VT]):
    """MutableMapping with waitable get and pop.

    TODO: exception support using outcome package
    """

    def __init__(self, *args, **kwargs):
        self._store = dict(*args, **kwargs)
        self._pending = {}  # key: Event

    def __getitem__(self, key):
        return self._store[key]

    async def get_wait(self, key: KT) -> VT:
        """Return value of given key, blocking until populated."""
        if key in self._store:
            return self._store[key]
        if key not in self._pending:
            self._pending[key] = trio.Event()
        await self._pending[key].wait()
        return self._store[key]

    async def pop_wait(self, key: KT) -> VT:
        """Remove key and return its value, blocking until populated."""
        value = await self.get_wait(key)
        del self._store[key]
        return value

    def is_waiting(self, key: KT) -> bool:
        """Return True if there is a task waiting for key."""
        return key in self._pending

    def __setitem__(self, key, value):
        self._store[key] = value
        if key in self._pending:
            self._pending.pop(key).set()

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __repr__(self):
        return repr(self._store)
